﻿using Azure.Cosmos;
using MeterTelemetry.Models;

namespace MeterTelemetry.Services;

public class TelemetryRepository
{
  private readonly IConfiguration _config;
  private readonly ILogger<TelemetryRepository> _logger;
  private readonly IHostEnvironment _enviroment;
  private const string DBNAME = "TelemetryDb";
  private const string CONTAINERNAME = "Telemetry";
  private CosmosClient _client;
  private bool _dbInitialized = false;

  public TelemetryRepository(IConfiguration config,
    ILogger<TelemetryRepository> logger,
    IHostEnvironment enviroment)
  {
    _config = config;
    _logger = logger;
    _enviroment = enviroment;
    var connString = _config.GetConnectionString("Cosmos");
    _client = new CosmosClient(connString);
  }

  public async Task<Telemetry> Get(string id, string monitorId)
  {
    try
    {
      var container = await GetContainer();
      var response = await container
              .ReadItemAsync<Telemetry>(
                id,
                new PartitionKey(monitorId));

      if (response.GetRawResponse().Status == 200)
      {
        return response.Value;
      }
      else
      {
        throw new Exception("Could not find the item");
      }
    }
    catch (Exception ex)
    {
      _logger.LogError($"Error: {ex}");
      throw;
    }
  }

  public async Task<Telemetry?> Get(string id)
  {
    try
    {
      var sql = $"SELECT * FROM c WHERE c.id = @id";

      var query = new QueryDefinition(sql)
        .WithParameter("@id", id);

      var container = await GetContainer();
      
      var iterator = container
              .GetItemQueryIterator<Telemetry>(query);

      var enumerator = iterator.GetAsyncEnumerator();

      if (await enumerator.MoveNextAsync())
      {
        return enumerator.Current;
      }
      else
      {
        // Not Found
        return null;
      }
    }
    catch (Exception ex)
    {
      _logger.LogError($"Error: {ex}");
      throw;
    }
  }


  public async Task<ICollection<Telemetry>> GetAll()
  {
    try
    {
      var container = await GetContainer();
      var query = new QueryDefinition($"SELECT * FROM c");
      var results = new List<Telemetry>();

      var iterator = container.GetItemQueryIterator<Telemetry>(query);
      await foreach (Telemetry result in iterator)
      {
        results.Add(result);
      }

      return results;
    }
    catch (Exception ex)
    {
      _logger.LogError($"Error: {ex}");
      throw;
    }
  }

  public async Task<IEnumerable<Telemetry>> GetFastWinds()
  {
    try
    {
      var container = await GetContainer();
      var query = new QueryDefinition(@"
SELECT VALUE c
FROM c
JOIN r IN c.Readings
WHERE r.WindSpeed > 10");
      var results = new List<Telemetry>();

      var iterator = container.GetItemQueryIterator<Telemetry>(query);
      await foreach (Telemetry result in iterator)
      {
        results.Add(result);
      }

      return results;
    }
    catch (Exception ex)
    {
      _logger.LogError($"Error: {ex}");
      throw;
    }
  }

  public async Task<IEnumerable<Telemetry>> GetTelemetryByMonitor(string monitorId)
  {
    try
    {
      var container = await GetContainer();

      var sql = @"SELECT VALUE c
                    FROM c
                    WHERE c.MonitorId = @monitorId";

      var query = new QueryDefinition(sql)
        .WithParameter("@monitorId", monitorId);

      var results = new List<Telemetry>();

      var iterator = container.GetItemQueryIterator<Telemetry>(query);
      await foreach (Telemetry result in iterator)
      {
        results.Add(result);
      }

      return results;
    }
    catch (Exception ex)
    {
      _logger.LogError($"Error: {ex}");
      throw;
    }
  }

  async Task InitializeDatabaseAsync()
  {
    if (_enviroment.IsDevelopment() && !_dbInitialized)
    {
      var result = await _client
        .CreateDatabaseIfNotExistsAsync(DBNAME);
      await result.Database
        .CreateContainerIfNotExistsAsync(
          CONTAINERNAME,
          "/MonitorId");

      _dbInitialized = true;
    }
  }

  public async Task<Telemetry> Create(Telemetry model)
  {
    try
    {
      var container = await GetContainer();
      var result = await container
        .CreateItemAsync(model,
                         new PartitionKey(model.MonitorId));
      return result.Value;
    }
    catch (Exception ex)
    {
      _logger.LogError($"Error: {ex}");
      throw;
    }
  }

  public async Task<Telemetry> Update(Telemetry model)
  {
    try
    {
      var container = await GetContainer();
      var result = await container.UpsertItemAsync(model, new PartitionKey(model.MonitorId));
      return result.Value;
    }
    catch (Exception ex)
    {
      _logger.LogError($"Error: {ex}");
      throw;
    }
  }

  public async Task<bool> Delete(string id)
  {
    try
    {
      var item = await Get(id);
      if (item is null) 
      {
        return false;
      }

      var container = await GetContainer();
      var result = await container.DeleteItemAsync<Telemetry>(item.Id, new PartitionKey(item.MonitorId));
      return true;
    }
    catch (Exception ex)
    {
      _logger.LogError($"Error: {ex}");
      return false;
    }
  }

  async Task<CosmosContainer> GetContainer()
  {
    await InitializeDatabaseAsync();
    return _client.GetContainer(DBNAME, CONTAINERNAME);
  }


}
